from django.test import TestCase

from pint import Context, DimensionalityError

import quantityfield.fields as fields
import quantityfield.helper as helper
from quantityfield.units import ureg


class TestMatchingUnitDimensionsHelper(TestCase):
    def test_valid_choices(self):
        helper.check_matching_unit_dimension(ureg, "meter", ["mile", "foot", "cm"])

    def test_invalid_choices(self):
        with self.assertRaises(DimensionalityError):
            helper.check_matching_unit_dimension(
                ureg, "meter", ["mile", "foot", "cm", "kg"]
            )


class TestEdgeCases(TestCase):
    def test_fix_unit_registry(self):
        field = fields.IntegerQuantityField("meter")
        with self.assertRaises(ValueError):
            field.fix_unit_registry(1)

    def test_get_prep_value(self):
        field = fields.IntegerQuantityField("meter")
        with self.assertRaises(ValueError):
            field.get_prep_value("foobar")


class TestContextHandling(TestCase):
    """Class to test ureg.context for compatible units as described in issue #99.
    pint allows users to define a special context for unit conversions,
    e.g. on earth a mass can directly be converted to a force given the
    acceleration 'constant' on earth.
    We will test the unit compatibility via the helper function for both
    both context activated globally via ureg and within a with block. Finally
    test the conversion integrated inside an IntegerQuantityField. Also the
    negatives are tested: without the context, mass should not be accepted
    as a matching unit for force.
    """

    def setUp(self):
        """Setup a pint context in the UnitRegistry."""
        # Define a context where mass is equated to force via earth's
        # standard acceleration of gravity and vice versa
        # (https://en.wikipedia.org/wiki/Standard_gravity)
        self.context = Context("earth")
        # mass -> force
        self.context.add_transformation(
            "[mass]", "[force]", lambda ureg, x: x * ureg.gravity
        )
        # force -> mass
        self.context.add_transformation(
            "[force]", "[mass]", lambda ureg, x: x / ureg.gravity
        )
        ureg.add_context(self.context)

    def test_context_global(self):
        """Activate ureg.context globally and test conversion compatibility directly."""
        # Activate context globally and test
        ureg.enable_contexts("earth")
        helper.check_matching_unit_dimension(ureg, "kg", ["newton", "kN", "ton"])
        ureg.disable_contexts()

    def test_context_with_block(self):
        """Activate ureg.context in with block and test conversion compatibility
        directly."""
        # Use context with the 'with' statement
        with ureg.context("earth"):
            helper.check_matching_unit_dimension(ureg, "kg", ["newton", "kN", "ton"])

    def test_invalid_context(self):
        """Negative test: Conversion mass to force should fail without context."""
        with self.assertRaises(DimensionalityError):
            helper.check_matching_unit_dimension(ureg, "kg", ["newton", "kN", "ton"])

    def test_field_w_context_global(self):
        """Negative test: Conversion mass to force should fail without context."""
        ureg.enable_contexts("earth")
        self.field = fields.IntegerQuantityField(
            base_units="kg", unit_choices=["newton", "kN", "ton"]
        )
        ureg.disable_contexts()

    def test_field_w_context_block(self):
        """Activate ureg.context globally and test conversion compatibility complete
        Field."""
        with ureg.context("earth"):
            self.field = fields.IntegerQuantityField(
                base_units="kg", unit_choices=["newton", "kN", "ton"]
            )

    def test_invalid_field_wo_context(self):
        """Negative test: Conversion mass to force should fail without context."""
        with self.assertRaises(DimensionalityError):
            self.field = fields.IntegerQuantityField(
                base_units="kg", unit_choices=["newton", "kN", "ton"]
            )

    def tearDown(self):
        """Disable and remove the contexts to not interfere with other tests."""
        # Clean up by disabling and removing the context
        ureg.disable_contexts()
        ureg.remove_context("earth")



from decimal import Decimal

import pytest

from django.db.models import Min, Subquery
from django.test import TestCase

from quantityfield.units import ureg
from tests.dummyapp.models import (
    BigIntFieldSaveModel,
    DecimalFieldSaveModel,
    EmptyHayBalePositiveInt,
    FloatFieldSaveModel,
    IntFieldSaveModel,
)

Quantity = ureg.Quantity


class BaseMixinQuantityFieldORM:
    """Base mixin for ORM tests for QuantityField types."""

    MODEL: type
    FIELD_NAME: str = "weight"
    CREATE_KWARGS_LIGHT: dict = {"name": "light", FIELD_NAME: 100}
    CREATE_KWARGS_HEAVY: dict = {"name": "heavy", FIELD_NAME: 200}
    EXPECTED_TYPE: type

    def setUp(self):
        self.light = self.MODEL.objects.create(**self.CREATE_KWARGS_LIGHT)
        self.heavy = self.MODEL.objects.create(**self.CREATE_KWARGS_HEAVY)

    def tearDown(self):
        self.MODEL.objects.all().delete()

    def test_bulk_update_with_subquery(self):
        min_value_qs = self.MODEL.objects.annotate(
            min_value=Min(self.FIELD_NAME)
        ).values("min_value")[:1]

        self.MODEL.objects.all().update(**{self.FIELD_NAME: Subquery(min_value_qs)})

        self.light.refresh_from_db()
        self.heavy.refresh_from_db()

        self.assertEqual(
            Quantity(self.EXPECTED_TYPE(100) * ureg.gram),
            getattr(self.light, self.FIELD_NAME),
        )
        self.assertEqual(
            Quantity(self.EXPECTED_TYPE(100) * ureg.gram),
            getattr(self.heavy, self.FIELD_NAME),
        )


@pytest.mark.django_db
class TestDecimalQuantityFieldORM(BaseMixinQuantityFieldORM, TestCase):
    MODEL = DecimalFieldSaveModel
    EXPECTED_TYPE = Decimal


@pytest.mark.django_db
class TestFloatQuantityFieldORM(BaseMixinQuantityFieldORM, TestCase):
    MODEL = FloatFieldSaveModel
    EXPECTED_TYPE = float


@pytest.mark.django_db
class TestIntegerQuantityFieldORM(BaseMixinQuantityFieldORM, TestCase):
    MODEL = IntFieldSaveModel
    EXPECTED_TYPE = int


@pytest.mark.django_db
class TestBigIntegerQuantityFieldORM(BaseMixinQuantityFieldORM, TestCase):
    MODEL = BigIntFieldSaveModel
    EXPECTED_TYPE = int


@pytest.mark.django_db
class TestPositiveIntegerQuantityFieldORM(BaseMixinQuantityFieldORM, TestCase):
    MODEL = EmptyHayBalePositiveInt
    EXPECTED_TYPE = int



from pint import DimensionalityError, UnitRegistry


def check_matching_unit_dimension(
    ureg: UnitRegistry, base_units: str, units_to_check: list[str]
) -> None:
    """
    Check if all units_to_check have the same Dimension like the base_units
    If not
    :raise DimensionalityError
    """
    base_unit = getattr(ureg, base_units)
    # create a pint quantity by multiplying unit with magnitude of 1
    base_quant = 1 * base_unit

    for unit_string in units_to_check:
        unit = getattr(ureg, unit_string)
        # try to convert base qunatity to new unit, this also work for ureg.context
        try:
            base_quant.to(unit)
        except DimensionalityError as e:
            raise DimensionalityError(base_unit, unit) from e



from django.contrib import admin

from . import models


class ReadOnlyEditing(admin.ModelAdmin):
    def get_readonly_fields(self, request, obj=None):
        if obj is not None:
            return list(self.get_fields(request))
        return []


admin.site.register(models.BigIntFieldSaveModel, ReadOnlyEditing)
admin.site.register(models.ChoicesDefinedInModel, ReadOnlyEditing)
admin.site.register(models.ChoicesDefinedInModelInt, ReadOnlyEditing)
admin.site.register(models.CustomUregDecimalHayBale, ReadOnlyEditing)
admin.site.register(models.CustomUregHayBale, ReadOnlyEditing)
admin.site.register(models.DecimalFieldSaveModel, ReadOnlyEditing)
admin.site.register(models.EmptyHayBaleBigInt, ReadOnlyEditing)
admin.site.register(models.EmptyHayBaleDecimal, ReadOnlyEditing)
admin.site.register(models.EmptyHayBaleFloat, ReadOnlyEditing)
admin.site.register(models.EmptyHayBaleInt, ReadOnlyEditing)
admin.site.register(models.FloatFieldSaveModel, ReadOnlyEditing)
admin.site.register(models.HayBale, ReadOnlyEditing)
admin.site.register(models.IntFieldSaveModel, ReadOnlyEditing)
admin.site.register(models.OffsetUnitFloatFieldSaveModel, ReadOnlyEditing)



from importlib.metadata import PackageNotFoundError, version

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "django-pint"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from quantityfield import fields, helper, settings, units, widgets

__all__ = ["fields", "helper", "settings", "units", "widgets"]
